import os.path
from autobean.xcheck import plugin
import autobean.utils.plugin_test_utils as utils


@utils.generate_tests(os.path.dirname(__file__), plugin)
def test():
    pass


def test_include_option():
    test_path = os.path.join(os.path.dirname(__file__), 'pass')
    source_path = os.path.join(test_path, 'source.bean')
    statement_path = os.path.join(test_path, '_statement.bean')
    entries, errors, options = utils.load_ledger(source_path)

    # assumptions
    assert not errors
    assert 'include' in options
    assert statement_path not in options['include']

    entries, errors = utils.apply_plugin(plugin, entries, options, None)

    assert not errors, errors
    assert statement_path in options['include'], 'statement files not added into options["include"]'



from collections import namedtuple
from decimal import Decimal
from typing import List, Optional
from beancount.core import amount
from beancount.core.data import Directive, Transaction, Balance, Pad
from autobean.utils.error_logger import ErrorLogger


InvalidDirectiveError = namedtuple('InvalidDirectiveError', 'source message entry')
DuplicatedOwnerError = namedtuple('InvalidDirectiveError', 'source message entry')


def select_viewpoint(entries: List[Directive], viewpoint: Optional[str], logger: ErrorLogger) -> List[Directive]:
    ret = []
    for entry in entries:
        if isinstance(entry, Balance) or isinstance(entry, Pad):
            continue
        if isinstance(entry, Transaction):
            entry = process_transaction(entry, viewpoint)
        if entry:
            ret.append(entry)
    return ret


def process_transaction(entry: Transaction, viewpoint: str) -> Optional[Transaction]:
    if viewpoint == 'everyone':
        return entry
    suffix = f':[{viewpoint}]'
    postings = []
    relevant = False
    for posting in entry.postings:
        if posting.account.startswith('[Residuals]:'):
            if not posting.account.endswith(suffix):
                postings.append(posting._replace(
                    units=amount.mul(posting.units, Decimal(-1)),
                ))
        else:
            if posting.account.endswith(suffix):
                postings.append(posting._replace(
                    account=posting.account.rsplit(':', 1)[0],
                ))
        if posting.account.endswith(suffix):
            relevant = True
    if postings:
        return entry._replace(
            flag=entry.flag if relevant else 'T',
            postings=postings,
        )



from typing import List, Dict, Tuple
from beancount.core.data import Directive
from beancount.ops import validation
from autobean.utils.error_logger import ErrorLogger
from autobean.share.include_context import include_context
from autobean.share.include import process_ledger
from autobean.share.fill_residuals import fill_residuals
from autobean.share.map_residual_accounts import map_residual_accounts
from autobean.share.open_subaccounts import open_subaccounts
from autobean.share.select_viewpoint import select_viewpoint


def plugin(entries: List[Directive], options: Dict, viewpoint: str) -> Tuple[List[Directive], List]:
    is_top_level = include_context['is_top_level']
    if not is_top_level:
        return entries, []
    include_context['is_top_level'] = False
    logger = ErrorLogger()
    errors = validation.validate(entries, options)
    logger.log_errors(errors)
    entries = process_ledger(entries, viewpoint == 'nobody', options, logger)
    if viewpoint != 'nobody':
        entries = fill_residuals(entries, options)
        entries = select_viewpoint(entries, viewpoint, logger)
        entries = map_residual_accounts(entries, logger)
        entries = open_subaccounts(entries, logger)
    include_context['is_top_level'] = True
    return entries, logger.errors



from typing import List, Set, DefaultDict
from collections import defaultdict
from beancount.core.data import Directive, Transaction, Open, Close
from autobean.utils.error_logger import ErrorLogger
from autobean.share import utils


def open_subaccounts(entries: List[Directive], logger: ErrorLogger) -> List[Directive]:
    subaccounts: DefaultDict[str, Set[str]] = defaultdict(set)
    for entry in entries:
        if not isinstance(entry, Transaction):
            continue
        for posting in entry.postings:
            if utils.is_subaccount(posting.account):
                subaccount = posting.account
                account = subaccount.rsplit(':', 1)[0]
                subaccounts[account].add(posting.account)
    ret = []
    for entry in entries:
        if not isinstance(entry, Open) and not isinstance(entry, Close):
            ret.append(entry)
            continue
        if not entry.account in subaccounts:
            ret.append(entry)
        for subaccount in subaccounts[entry.account]:
            ret.append(entry._replace(account=subaccount))
    return ret



from dataclasses import dataclass
from collections import namedtuple, defaultdict, deque, Counter
from datetime import timedelta
from decimal import Decimal
import functools
from typing import Dict, List, Iterable, Tuple, Optional
from beancount.core import amount
from beancount.core.data import Directive, Transaction, Custom, filter_txns, iter_entry_dates
from beancount.core.account import TYPE as ACCOUNT_TYPE
from autobean.utils.error_logger import ErrorLogger
from autobean.share import utils


InvalidDirectiveError = namedtuple('InvalidDirectiveError', 'source message entry')
UnresolvedLinkError = namedtuple('UnresolvedLinkError', 'source message entry')


@dataclass(repr=False, frozen=True)
class Link:
    directive: Custom

    @functools.cached_property
    def _values(self) -> Tuple[str, str, str, str]:
        return tuple(v.value for v in self.directive.values)

    @property
    def filename(self) -> str:
        return self._values[0]

    @property
    def account(self) -> str:
        return self._values[1]
    
    @property
    def complement_filename(self) -> str:
        return self._values[2]
    
    @property
    def complement_account(self) -> str:
        return self._values[3]

    def __repr__(self) -> str:
        return str(self._values)
    
    def valid(self) -> bool:
        if len(self.directive.values) != 4:
            return False
        if (tuple(v.dtype for v in self.directive.values) !=
                (str, ACCOUNT_TYPE, str, ACCOUNT_TYPE)):
            return False
        return True


def link_accounts(
        entries_by_file: Dict[str, List[Directive]],
        links: Iterable[Link],
        logger: ErrorLogger) -> List[Directive]:

    _check_links(entries_by_file, links, logger)
    edges = _build_graph(entries_by_file, links, logger)
    return _resolve_links(entries_by_file, edges, logger)


def _check_links(
        entries_by_file: Dict[str, List[Directive]],
        links: Iterable[Link],
        logger: ErrorLogger):
    all_endpoints = set()
    for link in links:
        if not link.valid():
            logger.log_error(InvalidDirectiveError(
                link.directive.meta,
                'autobean.share.link expects {filename} {account} '
                '{complement filename} {complement account} as arguments',
                link.directive,
            ))
            continue
        endpoints = [
            (link.filename, link.account),
            (link.complement_filename, link.complement_account),
        ]
        for ep in endpoints:
            if ep in all_endpoints:
                logger.log_error(InvalidDirectiveError(
                    link.directive.meta,
                    f'Account {ep[1]} in {ep[0]} has multiple links',
                    link.directive,
                ))
            if ep[0] not in entries_by_file:
                logger.log_error(InvalidDirectiveError(
                    link.directive.meta,
                    f'Ledger {ep[0]} was not included with '
                    f'"autobean.share.include" from this ledger',
                    link.directive,
                ))
            all_endpoints.add(ep)


def _build_graph(
        entries_by_file: Dict[str, List[Directive]],
        links: Iterable[Link],
        logger: ErrorLogger) -> Dict[int, List[Tuple[Transaction, str]]]:

    day = timedelta(days=1)
    edges = defaultdict(list) # id(txn) -> [(complement txn, account)]
    for link in links:
        if not link.valid():
            continue
        entries = entries_by_file.get(link.filename, [])
        complement_entries = entries_by_file.get(link.complement_filename, [])

        for entry in filter_txns(entries):
            expected_complement_feature = _transaction_feature(
                entry, link.account, True)
            if not expected_complement_feature[1]:
                continue
            # find complement txn and check duplicated complement
            complement_txn = None
            complement_duplicated = False
            for complement_entry in filter_txns(iter_entry_dates(
                    complement_entries, entry.date, entry.date + day)):
                complement_feature = _transaction_feature(
                    complement_entry, link.complement_account, False)
                if complement_feature == expected_complement_feature:
                    if not complement_txn:
                        complement_txn = complement_entry
                    else:
                        complement_duplicated = True
                        break
            # check duplicated
            duplicated = False
            found = False
            feature = _transaction_feature(entry, link.account, False)
            for entry2 in filter_txns(iter_entry_dates(
                    entries, entry.date, entry.date + day)):
                feature2 = _transaction_feature(entry2, link.account, False)
                if feature2 == feature:
                    if found:
                        duplicated = True
                        break
                    found = True
            if not complement_txn:
                logger.log_error(UnresolvedLinkError(
                    entry.meta,
                    f'No complement transaction found for link {link}',
                    entry,
                ))
            elif complement_duplicated:
                logger.log_error(UnresolvedLinkError(
                    entry.meta,
                    f'Multiple complement transactions found for link {link}',
                    entry,
                ))
            elif duplicated:
                logger.log_error(UnresolvedLinkError(
                    complement_txn.meta,
                    f'Multiple complement transactions found for link {link}',
                    complement_txn,
                ))
            else:
                edges[id(entry)].append((complement_txn, link.account))
                edges[id(complement_txn)].append((entry, link.complement_account))

        for complement_txn in filter_txns(complement_entries):
            complement_feature = _transaction_feature(
                complement_txn, link.complement_account, False)
            if complement_feature[1] and id(complement_txn) not in edges:
                logger.log_error(UnresolvedLinkError(
                    complement_txn.meta,
                    f'No complement transaction found for link {link}',
                    complement_txn,
                ))

    return edges


def _resolve_links(
        entries_by_file: Dict[str, List[Directive]],
        edges: Dict[int, List[Tuple[Transaction, str]]],
        logger: ErrorLogger) -> List[Directive]:
    ret = []
    all_visited = set()
    bad = set()
    for entries in entries_by_file.values():
        for entry in entries:
            if id(entry) in all_visited:
                continue
            if id(entry) not in edges or id(entry) in bad:
                ret.append(entry)
                continue

            q = deque([entry])
            visited = set()
            txns = []
            while q:
                u = q.popleft()
                if id(u) in visited:
                    continue
                visited.add(id(u))
                txns.append(u)
                for edge in edges[id(u)]:
                    q.append(edge[0])
            merged_txn = merge_transactions(txns, edges, logger)
            if merged_txn:
                ret.append(merged_txn)
                all_visited.update(visited)
            else:
                ret.append(entry)
                bad.update(visited)
    return ret


def _transaction_feature(
        entry: Transaction,
        account: str,
        negated: bool) -> Tuple[Optional[str], Counter]:
    link_key = entry.meta.get('share_link_key', None)

    posting_features = []
    for posting in entry.postings:
        if utils.main_account(posting.account) != account:
            continue
        if negated:
            units = amount.mul(posting.units, Decimal(-1))
        else:
            units = posting.units
        posting_features.append(units)
    return link_key, Counter(posting_features)


def merge_transactions(
        txns: List[Transaction],
        edges: Dict[int, Tuple[Transaction, str]],
        logger: ErrorLogger) -> Optional[Transaction]:
    date = None
    flag = None
    payee = None
    narration = ''
    tags = set()
    links = set()
    meta = {}
    postings = []
    compatible = True
    last_txn = None
    for txn in txns:
        last_txn = txn
        if date and txn.date and txn.date != date:
            compatible = False
        date = date or txn.date
        if flag and txn.flag and txn.flag != flag:
            compatible = False
        flag = flag or txn.flag
        if payee and txn.payee and txn.payee != payee:
            compatible = False
        payee = payee or txn.payee
        if narration and txn.narration and txn.narration != narration:
            compatible = False
        narration = narration or txn.narration
        for k, v in txn.meta.items():
            mv = meta.get(k, None)
            meta[k] = mv or v
            if mv and mv != v and k not in ('filename', 'lineno'):
                compatible = False
                break
        if not compatible:
            break
        accounts_to_remove = set(edge[1] for edge in edges[id(txn)])
        for posting in txn.postings:
            if utils.main_account(posting.account) not in accounts_to_remove:
                postings.append(posting)

    if not compatible:
        logger.log_error(UnresolvedLinkError(
            last_txn.meta,
            'Transaction and its complement do not agree on flag, payee, '
            'narration or meta',
            last_txn,
        ))
        return None
    txn = Transaction(
        date=date,
        meta=meta,
        flag=flag,
        payee=payee,
        narration=narration,
        tags=tags,
        links=links,
        postings=postings,
    )
    return txn



import re
from typing import List
from beancount.core.data import Directive, Transaction
from autobean.utils.error_logger import ErrorLogger


RESIDUAL_ACCOUNT = 'Assets:Receivable'
ERROR_ACCOUNT = 'Equity:Error'


def map_residual_accounts(entries: List[Directive], logger: ErrorLogger) -> List[Directive]:
    ret = []
    for entry in entries:
        if not isinstance(entry, Transaction):
            ret.append(entry)
            continue
        ret_postings = []
        for posting in entry.postings:
            account = posting.account
            account = re.sub(r'^\[Residuals\](?=:|$)', RESIDUAL_ACCOUNT.replace('\\', '\\\\'), account)
            account = re.sub(r'^\[Error\](?=:|$)', ERROR_ACCOUNT.replace('\\', '\\\\'), account)
            posting = posting._replace(
                account=account
            )
            ret_postings.append(posting)
        entry = entry._replace(
            postings=ret_postings,
        )
        ret.append(entry)
    return ret



from typing import List, Dict, Any
from collections import defaultdict
from beancount.core.data import Directive, Transaction
from beancount.core import interpolate
from autobean.share import utils


def fill_residuals(entries: List[Directive], options: Dict[str, Any]) -> List[Directive]:
    ret = []
    for entry in entries:
        if not isinstance(entry, Transaction):
            ret.append(entry)
            continue
        # Create residual postings for each party and add up all postings into real accounts
        # If the original transaction was not balanced we book residuals to error account
        # so we don't accidentially make it looks balanced for a subaccount in some cases.
        original_residual = interpolate.compute_residual(entry.postings)
        tolerances = interpolate.infer_tolerances(entry.postings, options)
        if original_residual.is_small(tolerances):
            subaccount_tmpl = '[Residuals]:[{}]'
        else:
            subaccount_tmpl = '[Error]:[{}]'
        postings = entry.postings[:]
        postings_by_party = defaultdict(list)
        for posting in entry.postings:
            party = posting.account.rsplit(':', 1)[1].strip('[]')
            postings_by_party[party].append(posting)
        for party, party_postings in postings_by_party.items():
            residual = interpolate.compute_residual(party_postings)
            # Use square brackets in account name to avoid collision with actual accounts
            subaccount = subaccount_tmpl.format(party)
            postings += utils.get_residual_postings(residual, subaccount)
        ret.append(entry._replace(postings=postings))
    return ret



from typing import List, Optional, Tuple, Dict, Callable, Union
import os
import sys
import copy
from collections import namedtuple
import pytest
from beancount import loader
from beancount.core.data import Directive, Transaction, Open, Close
from beancount.parser import printer
from autobean.utils.compare import compare_entries


Ledger = Tuple[List[Directive], List, Dict]
ExpectedError = namedtuple('ExpectedError', 'filename lineno message')
Expected = Union[Ledger, List[ExpectedError]]
TestcaseArg = namedtuple('Testcase', 'source plugin_arg expected_type expected')


def generate_tests(tests_path: str, plugin: Callable):
    ids, args = collect_testcases(tests_path)
    def decorator(func):

        def test(source: Ledger, plugin_arg: Optional[str], expected_type: str, expected: Expected):
            source_entries, source_errors, options = source
            assert not source_errors
            entries, errors = apply_plugin(plugin, source_entries, options, plugin_arg)
            entries = postprocess(entries)
            if expected_type == 'bean':
                assert not errors, errors
                assert_same_results(entries, expected)
            elif expected_type == 'errors':
                assert_same_errors(errors, expected)
            func()

        return pytest.mark.parametrize('source,plugin_arg,expected_type,expected', args, ids=ids)(test)
    return decorator


def collect_testcases(tests_path: str) -> Tuple[List[str], List[TestcaseArg]]:
    ids = []
    args = []
    suites = os.listdir(tests_path)
    for suite in suites:
        full_path = os.path.join(tests_path, suite)
        if not os.path.isdir(full_path) or suite.startswith('.') or suite.startswith('_'):
            continue
        source_path = os.path.join(full_path, 'source.bean')
        if not os.path.isfile(source_path):
            continue
        source = load_ledger(source_path)
        plugin_args = set()
        for filename in os.listdir(full_path):
            segs = filename.rsplit('.', 2)
            if len(segs) != 2:
                continue
            plugin_arg, expected_type = segs
            if plugin_arg == 'source' or plugin_arg.startswith('.') or plugin_arg.startswith('_'):
                continue
            if plugin_arg == 'None':
                plugin_arg = None
            output_path = os.path.join(full_path, filename)
            if expected_type == 'bean':  # expect results
                expected = load_ledger(output_path)[0]
            elif expected_type == 'errors':  # expect plugin errors
                expected = load_expected_errors(output_path)
            else:
                continue
            if plugin_arg in plugin_args:
                raise Exception(f'Multiple output files for "{plugin_arg}" in "{full_path}"')
            plugin_args.add(plugin_arg)
            ids.append(f'{suite} ({plugin_arg})')
            args.append(TestcaseArg(copy.deepcopy(source), plugin_arg, expected_type, expected))
    return ids, args


def load_ledger(path: str) -> Ledger:
    entries, errors, options = loader.load_file(path)
    return entries, errors, options


def load_expected_errors(path: str) -> List[ExpectedError]:
    with open(path) as f:
        errors = f.readlines()
    rets = []
    for error in errors:
        segs = error.strip().split(':', 2)
        rets.append(ExpectedError(segs[0], int(segs[1]), segs[2]))
    return rets


def assert_same_results(actuals: List[Directive], expecteds: List[Directive]):
    same, missings1, missings2 = compare_entries(actuals, expecteds)
    for missing in missings1:
        print('Unexpected entry:', file=sys.stderr)
        printer.print_entry(missing, file=sys.stderr)
    for missing in missings2:
        print('Missing entry:', file=sys.stderr)
        printer.print_entry(missing, file=sys.stderr)
    assert same


def assert_same_errors(actuals: List, expecteds: List[ExpectedError]):
    assert len(actuals) == len(expecteds)
    for actual, expected in zip(actuals, expecteds):
        assert os.path.basename(actual.source['filename']) == expected.filename
        assert actual.source['lineno'] == expected.lineno
        assert expected.message in actual.message


def apply_plugin(
        plugin: Callable,
        entries: List[Directive],
        options: Dict,
        plugin_arg: Optional[str]) -> Tuple[List[Directive], List]:

    if plugin_arg is None:
        return plugin(entries, options)
    return plugin(entries, options, plugin_arg)


def postprocess_account(account: str) -> str:
    """Replaces subaccount square brackets with hyphens.

    This is to allow testdata to be parsed normally.
    """
    return account.replace('[', 'TEST--').replace(']', '--')


def postprocess(entries: List[Directive]) -> List[Directive]:
    ret = []
    for entry in entries:
        if isinstance(entry, Transaction):
            postings = [
                posting._replace(
                    account=postprocess_account(posting.account))
                for posting in entry.postings
            ]
            ret.append(entry._replace(postings=postings))
        elif isinstance(entry, Open) or isinstance(entry, Close):
            ret.append(entry._replace(
                account=postprocess_account(entry.account)))
        else:
            ret.append(entry)

    return ret


def assert_results(self, testcase: str, plugin_options: Optional[List[Optional[str]]] = None):
    entries, options = self.load_source(testcase)
    if plugin_options is None:
        plugin_options = [None]
    for plugin_option in plugin_options:
        actual, errors = self.invoke_plugin(entries, options, plugin_option)
        assert not errors
        expected = self.load_results(testcase, plugin_option)
        self.assert_same_results(actual, expected)



from typing import List
from beancount.loader import LoadError
from beancount.core.data import Directive


class ErrorLogger:
    errors: List

    def __init__(self):
        self.errors = []

    def log_error(self, error):
        self.errors.append(error)

    def log_errors(self, errors: List):
        self.errors.extend(errors)

    def log_loading_errors(self, errors: List, entry: Directive):
        for error in errors:
            if isinstance(error, LoadError):
                error = error._replace(
                    source=entry.meta,
                    entry=entry,
                )
            self.errors.append(error)



from typing import List
from collections import Counter
from beancount.core import data
from beancount.core.data import Directive
from beancount.core.compare import hash_entry


# This is similar to beancount.core.comparer.compare_entries but allows duplicated entries
def compare_entries(entries1: List[Directive], entries2: List[Directive]):
    hashes1 = {
        hash_entry(entry, exclude_meta=True): entry for entry in entries1}
    hashes2 = {
        hash_entry(entry, exclude_meta=True): entry for entry in entries2}

    keys1 = Counter(hash_entry(entry, exclude_meta=True) for entry in entries1)
    keys2 = Counter(hash_entry(entry, exclude_meta=True) for entry in entries2)

    same = keys1 == keys2
    missing1 = data.sorted([hashes1[key] for key in keys1 - keys2])
    missing2 = data.sorted([hashes2[key] for key in keys2 - keys1])
    return same, missing1, missing2



from typing import List, Dict, Tuple, Set
from collections import namedtuple
import os.path
from beancount.core.data import Directive, Custom
from beancount import loader
from autobean.utils.plugin_base import PluginBase


def plugin(entries: List[Directive], options: Dict) -> Tuple[List[Directive], List]:
    plugin = IncludePlugin()
    return plugin.process(entries, options)


InvalidDirectiveError = namedtuple('InvalidDirectiveError', 'source message entry')


class IncludePlugin(PluginBase):
    _includes: Set[str]

    def __init__(self):
        super().__init__()
        self._includes = set()

    def process(self, entries: List[Directive], options: Dict) -> Tuple[List[Directive], List]:
        self._includes = set(options['include'])
        ret = []
        for entry in entries:
            if isinstance(entry, Custom) and entry.type == 'autobean.include':
                ret.extend(self.process_include_directive(entry))
            else:
                ret.append(entry)
        # Allow tools to refresh data when included files are updated.
        options['include'] = list(self._includes)
        return ret, self._errors

    def process_include_directive(self, entry: Custom) -> List[Directive]:
        if len(entry.values) != 1:
            self._error(InvalidDirectiveError(
                entry.meta, 'autobean.include expects 1 argument but {} are given'.format(len(entry.values)), entry
            ))
            return []
        if entry.values[0].dtype is not str:
            self._error(InvalidDirectiveError(
                entry.meta, 'autobean.include expects a path as argument', entry
            ))
            return []
        path = entry.values[0].value
        path = os.path.join(os.path.dirname(entry.meta['filename']), path)
        entries, errors, _ = loader.load_file(path)
        self._loading_errors(errors, entry.meta, entry)
        self._includes.add(path)
        return entries



from typing import Iterator, Union
from decimal import Decimal
from beancount.core.data import Posting, Directive, Custom
from beancount.core.amount import Amount, mul as amount_mul, div as amount_div
from beancount.core.position import CostSpec
from beancount.core.inventory import Inventory


def amount_distrib(amount: Amount, weight: Decimal, total_weight: Decimal):
    return amount_div(amount_mul(amount, weight), total_weight)


def costspec_distrib(costspec: CostSpec, weight: Decimal, total_weight: Decimal):
    number_total = costspec.number_total
    if number_total is not None:
        amount = Amount(number_total, costspec.currency)
        number_total = amount_distrib(amount, weight, total_weight)
    return costspec._replace(
        number_total = number_total
    )


def posting_distrib(posting: Posting, weight: Decimal, total_weight: Decimal):
    units = amount_distrib(posting.units, weight, total_weight)
    cost = posting.cost
    if isinstance(cost, CostSpec):
        cost = costspec_distrib(cost, weight, total_weight)
    return posting._replace(
        units=units,
        cost=cost,
    )


def get_residual_postings(residual: Inventory, account_rounding: str):
    return [
        Posting(
            account=account_rounding,
            units=-position.units,
            cost=position.cost,
            price=None,
            flag=None,
            meta={}
        )
        for position in residual
    ]


def is_share_policy_directive(entry: Directive):
    return isinstance(entry, Custom) and entry.type == 'autobean.share.policy'


def is_proportionate_assertion_directive(entry: Directive):
    return isinstance(entry, Custom) and entry.type == 'autobean.share.proportionate'


def is_include_directive(entry: Directive):
    return isinstance(entry, Custom) and entry.type == 'autobean.share.include'


def is_autobean_share_directive(entry: Directive):
    return isinstance(entry, Custom) and (entry.type.startswith('autobean.share.') or entry.type == 'autobean.share')


def is_owner_directive(entry: Directive):
    return isinstance(entry, Custom) and entry.type == 'autobean.share.owner'


def is_subaccount(account: str):
    lastseg = account.rsplit(':', 1)[1]
    return lastseg.startswith('[') and lastseg.endswith(']')


def main_account(account: str):
    if is_subaccount(account):
        return account.rsplit(':', 1)[0]
    return account


def ancestor_accounts(account: str) -> Iterator[str]:
    segs = account.split(':')
    while segs:
        yield ':'.join(segs)
        segs.pop()


def strip_meta(entry: Union[Directive, Posting]):
    if not entry.meta:
        return entry
    meta = {
        k: v
        for k, v in entry.meta.items()
        if not k.startswith('share-') and k not in (
            'share_policy',
            'share_recursive',
            'share_prorata')
    }
    return entry._replace(meta=meta)



from typing import List, Dict, Tuple, Iterator, Set
from collections import namedtuple, Counter
import os.path
import datetime
from beancount.core.data import Directive, Custom, Transaction, Balance, Posting
from beancount import loader
from beancount.ops.validation import ValidationError
import beancount.core.account
import beancount.ops.balance
from autobean.utils.plugin_base import PluginBase


def plugin(entries: List[Directive], options: Dict) -> Tuple[List[Directive], List]:
    plugin = CrossCheckPlugin()
    return plugin.process(entries, options)


InvalidDirectiveError = namedtuple('InvalidDirectiveError', 'source message entry')
CrossCheckError = namedtuple('CrossCheckError', 'source message entry')


class PostingToCompare:
    posting: Posting
    transaction: Transaction

    def __init__(self, posting: Posting, transaction: Transaction):
        self.posting = posting
        self.transaction = transaction

    def __eq__(self, other):
        return self.transaction.date == other.transaction.date \
               and self.posting.account == other.posting.account \
               and self.posting.units == other.posting.units

    def __hash__(self):
        return hash((self.transaction.date, self.posting.account, self.posting.units))


class CrossCheckPlugin(PluginBase):
    _includes: Set[str]

    def __init__(self):
        super().__init__()
        self._includes = set()

    def process(self, entries: List[Directive], options: Dict) -> Tuple[List[Directive], List]:
        self._includes = set(options['include'])
        generated_entries = []
        for entry in entries:
            if isinstance(entry, Custom) and entry.type == 'autobean.xcheck':
                generated_entries += self.process_xcheck_directive(entry, entries)
        # Allow tools to refresh data when included files are updated.
        options['include'] = list(self._includes)
        return entries + generated_entries, self._errors

    def process_xcheck_directive(self, entry: Custom, entries: List[Directive]) -> List[Directive]:
        if len(entry.values) < 2:
            self._error(InvalidDirectiveError(
                entry.meta, 'autobean.xcheck expects at least 2 arguments but 1 is given', entry
            ))
            return []
        if entry.values[0].dtype is not str \
                or entry.values[1].dtype is not datetime.date \
                or any(arg.dtype != beancount.core.account.TYPE for arg in entry.values[2:]):
            self._error(InvalidDirectiveError(
                entry.meta, 'autobean.xcheck expects a path, a start date and zero or more accounts as arguments', entry
            ))
            return []
        path = entry.values[0].value
        path = os.path.join(os.path.dirname(entry.meta['filename']), path)
        start = entry.values[1].value
        end = entry.date
        accounts = {arg.value for arg in entry.values[2:]}
        stmt_entries, stmt_errors, _ = loader.load_file(path)
        stmt_errors = [error for error in stmt_errors if not isinstance(error, ValidationError)]
        if stmt_errors:
            self._loading_errors(stmt_errors, entry.meta, entry)
            return []

        entries = self.filter_by_time_period(entries, start, end)
        stmt_entries = self.filter_by_time_period(stmt_entries, start, end)
        postings = list(self.extract_related_postings(entries, accounts))
        stmt_postings = list(self.extract_related_postings(stmt_entries, accounts))

        _, unexpected, missing = self.compare_postings(postings, stmt_postings)
        for posting in unexpected:
            self._error(CrossCheckError(
                posting.posting.meta, 'Unexpected posting', posting.transaction
            ))
        for posting in missing:
            self._error(CrossCheckError(
                posting.posting.meta, 'Missing posting', posting.transaction
            ))
        self._includes.add(path)
        return [entry for entry in stmt_entries if isinstance(entry, Balance)]

    @staticmethod
    def extract_related_postings(entries: List[Directive], accounts: Set[str]) -> Iterator[PostingToCompare]:
        for entry in entries:
            if not isinstance(entry, Transaction):
                continue
            for posting in entry.postings:
                if not accounts or posting.account in accounts:
                    yield PostingToCompare(posting, entry)

    @classmethod
    def compare_postings(cls, postings1: List[PostingToCompare], postings2: List[PostingToCompare]) -> Tuple[bool, List[PostingToCompare], List[PostingToCompare]]:
        missings2 = cls.find_missings(postings1, postings2)
        missings1 = cls.find_missings(postings2, postings1)
        same = not missings1 and not missings2
        return same, missings1, missings2

    @staticmethod
    def find_missings(postings1: List[PostingToCompare], postings2: List[PostingToCompare]) -> List[PostingToCompare]:
        hashed1 = Counter(postings1)
        missings2 = []
        for posting in postings2:
            if hashed1[posting]:
                hashed1[posting] -= 1
            else:
                missings2.append(posting)
        return missings2

    @staticmethod
    def filter_by_time_period(entries: List[Directive], start: datetime.date, end: datetime.date):
        return [entry for entry in entries if start <= entry.date < end]



from typing import List
from beancount.loader import LoadError
from beancount.core.data import Meta, Directive


class PluginBase:
    _errors: List

    def __init__(self):
        self._errors = []

    def _error(self, error):
        self._errors.append(error)

    def _loading_errors(self, errors: List, meta: Meta, entry: Directive):
        for error in errors:
            if isinstance(error, LoadError):
                error = error._replace(source=meta, entry=entry)
            self._errors.append(error)



from typing import List, Dict, Tuple, Set, Any, Optional
from collections import namedtuple
import os.path
from beancount.core.data import Directive, Transaction, Custom, Open
from beancount import loader
from autobean.utils.error_logger import ErrorLogger
from autobean.share import utils
from autobean.share.split_postings import split_postings
from autobean.share.link_accounts import link_accounts, Link


InvalidDirectiveError = namedtuple('InvalidDirectiveError', 'source message entry')


def process_ledger(entries: List[Directive], is_nobody: bool, options: Dict[str, Any], logger: ErrorLogger) -> List[Directive]:
    # For nobody viewpoint, we still fill the residuals but ignore the
    # results so proportionate assertion works.
    filled_entries = split_postings(entries, logger)
    if not is_nobody:
        entries = filled_entries
    else:
        entries = filter_out_share_meta(entries)
    includes = set(options['include'])
    entries = process_included_files(entries, is_nobody, includes, logger)
    entries = filter_out_share_directives(entries)
    # Allow tools to refresh data when included files are updated
    options['include'][:] = list(includes)
    return entries


def process_included_files(entries: List[Directive], is_nobody: bool, includes: Set[str], logger: ErrorLogger) -> List[Directive]:
    ret = []
    links = []
    entries_by_file = {}
    for entry in entries:
        if isinstance(entry, Custom) and entry.type == 'autobean.share.include':
            filename, file_entries = process_include_directive(entry, is_nobody, includes, logger)
            if filename:
                entries_by_file[filename] = file_entries
        if isinstance(entry, Custom) and entry.type == 'autobean.share.link':
            links.append(Link(entry))
        else:
            ret.append(entry)
    resolved_entries = link_accounts(entries_by_file, links, logger)
    ret.extend(resolved_entries)
    ret = deduplicate_opens(ret)
    return ret


def process_include_directive(
        entry: Custom,
        is_nobody: bool,
        includes: Set[str],
        logger: ErrorLogger) -> Tuple[Optional[str], List[Directive]]:
    if len(entry.values) != 1:
        logger.log_error(InvalidDirectiveError(
            entry.meta, 'autobean.share.include expects 1 argument but {} are given'.format(len(entry.values)), entry
        ))
        return None, []
    if entry.values[0].dtype is not str:
        logger.log_error(InvalidDirectiveError(
            entry.meta, 'autobean.share.include expects a path as argument', entry
        ))
        return None, []
    filename = entry.values[0].value
    path = os.path.join(os.path.dirname(entry.meta['filename']), filename)
    entries, errors, options = loader.load_file(path)
    logger.log_loading_errors(errors, entry)
    entries = process_ledger(entries, is_nobody, options, logger)
    includes.update(set(options['include']))
    return filename, entries


def filter_out_share_directives(entries: List[Directive]) -> List[Directive]:
    return [
        entry
        for entry in entries
        if not utils.is_autobean_share_directive(entry)
    ]

def filter_out_share_meta(entries: List[Directive]) -> List[Directive]:
    ret = []
    for entry in entries:
        entry = utils.strip_meta(entry)
        if isinstance(entry, Transaction):
            entry.postings[:] = [
                utils.strip_meta(posting)
                for posting in entry.postings
            ]
        ret.append(entry)
    return ret
    

def deduplicate_opens(entries: List[Directive]) -> List[Directive]:
    ret = []
    open_accounts = set()
    for entry in entries:
        if not isinstance(entry, Open):
            ret.append(entry)
        elif entry.account not in open_accounts:
            open_accounts.add(entry.account)
            ret.append(entry)
    return ret



from decimal import Decimal


class Policy(dict):
    total_weight: Decimal

    def replace(self, new_policy):
        super().clear()
        super().update(new_policy)
        self.total_weight = new_policy.total_weight



from typing import List, Dict, Optional, Tuple, Set
from collections import namedtuple, defaultdict
from decimal import Decimal
from beancount.core.data import Directive, Open, Close, Custom, Transaction, Posting
from beancount.core import realization, inventory
import beancount.core.account
from autobean.utils.error_logger import ErrorLogger
from autobean.share.policy import Policy
from autobean.share import utils


def split_postings(entries: List[Directive], logger: ErrorLogger) -> List[Directive]:
    plugin = SplitPostingsPlugin(logger)
    return plugin.process(entries)


AccountNotFoundError = namedtuple('AccountNotFoundError', 'source message entry')
NoApplicablePolicyError = namedtuple('NoApplicablePolicyError', 'source message entry')
InvalidDirectiveError = namedtuple('InvalidDirectiveError', 'source message entry')
InvalidSharePolicyError = namedtuple('InvalidSharePolicyError', 'source message entry')
ProportionateError = namedtuple('ProportionateError', 'source message entry')
TOLERANCE = Decimal(1e-6)


class SplitPostingsPlugin:
    logger: ErrorLogger
    named_policies: Dict[str, Policy]
    account_policies: Dict[Tuple[str, bool], Policy]  # (account, recursive) -> policy
    real_root: realization.RealAccount
    open_accounts: Set[str]

    def __init__(self, logger: ErrorLogger):
        self.logger = logger
        self.named_policies = defaultdict(Policy)
        self.account_policies = defaultdict(Policy)
        self.real_root = realization.RealAccount('')
        self.open_accounts = set()
        super().__init__()

    def process(self, entries: List[Directive]) -> List[Directive]:
        ret = []
        for entry in entries:
            if utils.is_share_policy_directive(entry):
                self.process_share_policy_definition(entry)
                entry = None
            elif utils.is_proportionate_assertion_directive(entry):
                self.process_proportionate(entry)
            elif isinstance(entry, Open):
                entry = self.process_open(entry)
            elif isinstance(entry, Close):
                self.process_close(entry)
            elif isinstance(entry, Transaction):
                entry = self.process_transaction(entry)
            if entry:
                ret.append(entry)
        return ret

    def process_share_policy_definition(self, entry: Custom):
        if len(entry.values) != 1:
            self.logger.log_error(InvalidDirectiveError(
                entry.meta, 'Share policy definition expects 1 account or string argument but {} are given'.format(len(entry.values)), entry
            ))
        elif entry.values[0].dtype == beancount.core.account.TYPE:
            # account policy
            account = entry.values[0].value
            recursive = entry.meta.get('share_recursive', True)
            if self.get_referenced_accounts(account, recursive, entry):
                policy = self.extract_policy(entry.meta, entry)
                self.account_policies[(account, recursive)].replace(policy)
        elif entry.values[0].dtype is str:
            # named policy
            policy_name = entry.values[0].value
            if not policy_name:
                self.logger.log_error(InvalidSharePolicyError(
                    entry.meta, 'Policy name cannot be empty', entry
                ))
                return
            policy = self.extract_policy(entry.meta, entry)
            # Allow changes on policies to propagate
            if 'share_policy' in entry.meta:
                self.named_policies[policy_name] = policy
            else:
                self.named_policies[policy_name].replace(policy)
        else:
            self.logger.log_error(InvalidSharePolicyError(
                entry.meta, 'Share policy must be assigned a name or applied on an account', entry
            ))

    def process_open(self, entry: Open) -> Open:
        self.open_accounts.add(entry.account)
        policy = self.extract_policy(entry.meta, entry, required=False)
        if policy:
            recursive = entry.meta.get('share_recursive', False)
            self.account_policies[(entry.account, recursive)] = policy
        open = utils.strip_meta(entry)
        return open

    def process_close(self, entry: Close):
        self.open_accounts.remove(entry.account)
    
    def prorata_selector(self, posting: Posting) -> Optional[Tuple[int, str]]:
        return (posting.account.split(':', 1)[0], posting.units.currency)

    def process_transaction(self, entry: Transaction) -> Transaction:
        transaction_policy = self.extract_policy(entry.meta, entry, required=False)
        # Split postings proportionately by party
        postings = []
        prorata_policies = {}
        for posting in entry.postings:
            prorata_selector = self.prorata_selector(posting)
            if posting.meta and posting.meta.get('share_prorata', False) == True:
                continue
            policy = self.get_posting_policy(posting, entry, transaction_policy, ignore_error=True)
            prorata_policy = prorata_policies.get(prorata_selector, None)
            if not prorata_policy:
                prorata_policy = Policy()
                prorata_policy.total_weight = Decimal(0)
                prorata_policies[prorata_selector] = prorata_policy
            for party, share in policy.items():
                if share:
                    split_amount = utils.amount_distrib(posting.units, share, policy.total_weight)
                    prorata_policy.setdefault(party, Decimal(0))
                    prorata_policy[party] += split_amount.number
                    prorata_policy.total_weight += split_amount.number

        for posting in entry.postings:
            if posting.meta and posting.meta.get('share_prorata', False) == True:
                policy = prorata_policies[self.prorata_selector(posting)]
            else:
                policy = self.get_posting_policy(posting, entry, transaction_policy)
            for party, share in policy.items():
                if share:
                    subaccount = posting.account + ':[{}]'.format(party)
                    split_posting = utils.posting_distrib(posting, share, policy.total_weight)
                    split_posting = utils.strip_meta(split_posting)
                    split_posting = split_posting._replace(
                        account=subaccount,
                    )
                    postings.append(split_posting)
        # Finalize processed postings
        entry = utils.strip_meta(entry)
        entry = entry._replace(
            postings=postings,
        )
        self.realize_transaction(entry)
        return entry

    def process_proportionate(self, entry: Custom):
        if len(entry.values) != 1:
            self.logger.log_error(InvalidDirectiveError(
                entry.meta, 'Proportionate assertion expects 1 account argument but {} are given'.format(len(entry.values)), entry
            ))
            return
        if entry.values[0].dtype != beancount.core.account.TYPE:
            self.logger.log_error(InvalidDirectiveError(
                entry.meta, 'Proportionate assertion must be applied on an account', entry
            ))
            return
        account = entry.values[0].value
        recursive = entry.meta.get('share_recursive', True)
        accounts = self.get_referenced_accounts(account, recursive, entry)
        for account in accounts:
            policy = self.extract_policy(entry.meta, entry, required=False) \
                or self.get_account_policy(account) \
                or self.named_policies.get('default', None)
            if not policy:
                self.logger.log_error(NoApplicablePolicyError(
                    entry.meta, 'No applicable share policy to proportionate assertion on account "{}"'.format(account), entry
                ))
                continue
            reference_inventory = None
            real_accounts = {
                k[1:-1]: v
                for k, v in realization.get(self.real_root, account, {}).items()
                if k[0] == '[' and k[-1] == ']'
            }
            for party, share in policy.items():
                real_account = real_accounts.get(party, None)
                if real_account is None:
                    balance = inventory.Inventory()
                else:
                    balance = real_account.balance
                if share == 0:
                    diff = balance
                else:
                    normalized_inventory = balance * (1 / share)
                    if reference_inventory is None:
                        reference_inventory = normalized_inventory
                        continue
                    else:
                        diff = abs(normalized_inventory + (-reference_inventory))
                if not diff.is_small(TOLERANCE):
                    self.logger.log_error(ProportionateError(
                        entry.meta, 'Account "{}" is disproportionate'.format(account), entry
                    ))
            # parties not in policy should have zero balance
            for party in set(real_accounts.keys()) - set(policy.keys()):
                real_account = real_accounts[party]
                if not real_account.balance.is_small(TOLERANCE):
                    self.logger.log_error(ProportionateError(
                        entry.meta, 'Account "{}" is disproportionate'.format(account), entry
                    ))

    def realize_transaction(self, entry: Transaction):
        for posting in entry.postings:
            real_account = realization.get_or_create(self.real_root, posting.account)
            real_account.balance.add_position(posting)

    def get_referenced_accounts(self, account: str, recursive: bool, entry: Directive):
        if recursive:
            accounts = [
                open_account
                for open_account in self.open_accounts
                if open_account == account or open_account.startswith(account + ':')
            ]
            if not accounts:
                self.logger.log_error(AccountNotFoundError(
                    entry.meta, 'Account "{}" is not found or inactive, as well as all its sub-accounts'.format(account), entry
                ))
        else:
            accounts = [account] if account in self.open_accounts else []
            if not accounts:
                self.logger.log_error(AccountNotFoundError(
                    entry.meta, 'Account "{}" is not found or inactive'.format(account), entry
                ))
        return accounts

    def get_account_policy(self, account: str) -> Optional[Policy]:
        policy = self.account_policies.get((account, False))
        if policy:
            return policy
        for account in utils.ancestor_accounts(account):
            policy = self.account_policies.get((account, True), None)
            if policy:
                return policy

    def get_posting_policy(self, posting: Posting, transaction: Transaction, transaction_policy: Optional[Policy], ignore_error: bool = False) -> Policy:
        policy = posting.meta and self.extract_policy(posting.meta, transaction, required=False) \
            or self.get_account_policy(posting.account) \
            or transaction_policy \
            or self.named_policies.get('default', None)
        if not policy:
            if not ignore_error:
                self.logger.log_error(NoApplicablePolicyError(
                    posting.meta or transaction.meta, 'No applicable share policy to account "{}"'.format(posting.account), transaction
                ))
            policy = Policy()
            policy.total_weight = 0
        return policy

    def extract_policy(self, meta: Dict, entry: Optional[Directive] = None, required: bool = True) -> Optional[Policy]:
        policy = Policy()
        total_weight = 0
        found = False
        for k, v in meta.items():
            if k.startswith('share-'):
                party_name = k[len('share-'):]
                if not party_name:
                    self.logger.log_error(InvalidSharePolicyError(
                        meta, 'Expect a name to share with', entry
                    ))
                elif not isinstance(v, Decimal):
                    self.logger.log_error(InvalidSharePolicyError(
                        meta, 'Expect a weight of share in decimal', entry
                    ))
                elif not party_name[0].isupper():
                    self.logger.log_error(InvalidSharePolicyError(
                        meta, 'The name to share with must start with an uppercase letter', entry
                    ))
                else:
                    found = True
                    policy[party_name] = v
                    total_weight += v
        policy.total_weight = total_weight
        if 'share_policy' in meta:
            if found:
                self.logger.log_error(InvalidSharePolicyError(
                    meta, 'Mixed use of "share_policy" and "share-" attributes', entry
                ))
            elif meta['share_policy'] not in self.named_policies:
                self.logger.log_error(InvalidSharePolicyError(
                    meta, 'Share policy "{}" does not exist'.format(meta['share_policy']), entry
                ))
            else:
                found = True
                policy = self.named_policies[meta['share_policy']]
        if required and not found:
            self.logger.log_error(InvalidSharePolicyError(
                meta, 'Expect definition of share policy', entry
            ))
        return policy if found or required else None
